=== modified file 'breezy/branch.py'
--- breezy/branch.py	2020-01-18 15:43:04 +0000
+++ breezy/branch.py	2020-01-19 03:22:04 +0000
@@ -51,7 +51,7 @@
     text_type,
     viewitems,
     )
-from .trace import mutter, mutter_callsite, note, is_quiet
+from .trace import mutter, mutter_callsite, note, is_quiet, warning
 
 
 class UnstackableBranchFormat(errors.BzrError):
@@ -1273,18 +1273,7 @@
     def update_references(self, target):
         if not getattr(self._format, 'supports_reference_locations', False):
             return
-        reference_dict = self._get_all_reference_info()
-        if len(reference_dict) == 0:
-            return
-        old_base = self.base
-        new_base = target.base
-        target_reference_dict = target._get_all_reference_info()
-        for tree_path, (branch_location, file_id) in viewitems(reference_dict):
-            branch_location = urlutils.rebase_url(branch_location,
-                                                  old_base, new_base)
-            target_reference_dict.setdefault(
-                tree_path, (branch_location, file_id))
-        target._set_all_reference_info(target_reference_dict)
+        return InterBranch.get(self, target).update_references()
 
     def check(self, refs):
         """Check consistency of the branch.
@@ -1405,6 +1394,9 @@
         with basis_tree.lock_read():
             for path in basis_tree.iter_references():
                 reference_parent = tree.reference_parent(path)
+                if reference_parent is None:
+                    warning('Branch location for %s unknown.', path)
+                    continue
                 reference_parent.create_checkout(
                     tree.abspath(path),
                     basis_tree.get_reference_revision(path), lightweight)
@@ -2101,6 +2093,11 @@
         """
         raise NotImplementedError(self.fetch)
 
+    def update_references(self):
+        """Import reference information from source to target.
+        """
+        raise NotImplementedError(self.update_references)
+
 
 def _fix_overwrite_type(overwrite):
     if isinstance(overwrite, bool):
@@ -2137,8 +2134,8 @@
                      be truncated to end with revision_id.
         """
         with self.source.lock_read(), self.target.lock_write():
-            self.source.update_references(self.target)
             self.source._synchronize_history(self.target, revision_id)
+            self.update_references()
             try:
                 parent = self.source.get_parent()
             except errors.InaccessibleParent as e:
@@ -2307,7 +2304,6 @@
         result.source_branch = self.source
         result.target_branch = self.target
         result.old_revno, result.old_revid = self.target.last_revision_info()
-        self.source.update_references(self.target)
         overwrite = _fix_overwrite_type(overwrite)
         if result.old_revid != stop_revision:
             # We assume that during 'push' this repository is closer than
@@ -2319,6 +2315,7 @@
             result.tag_updates, result.tag_conflicts = (
                 self.source.tags.merge_to(
                     self.target.tags, "tags" in overwrite))
+        self.update_references()
         result.new_revno, result.new_revid = self.target.last_revision_info()
         return result
 
@@ -2354,7 +2351,6 @@
         with self.source.lock_read():
             # We assume that during 'pull' the target repository is closer than
             # the source one.
-            self.source.update_references(self.target)
             graph = self.target.repository.get_graph(self.source.repository)
             # TODO: Branch formats should have a flag that indicates
             # that revno's are expensive, and pull() should honor that flag.
@@ -2371,6 +2367,7 @@
                 self.source.tags.merge_to(
                     self.target.tags, "tags" in overwrite,
                     ignore_master=not merge_tags_to_master))
+            self.update_references()
             result.new_revno, result.new_revid = (
                 self.target.last_revision_info())
             if _hook_master:
@@ -2384,5 +2381,21 @@
                     hook(result)
             return result
 
+    def update_references(self):
+        if not getattr(self.source._format, 'supports_reference_locations', False):
+            return
+        reference_dict = self.source._get_all_reference_info()
+        if len(reference_dict) == 0:
+            return
+        old_base = self.source.base
+        new_base = self.target.base
+        target_reference_dict = self.target._get_all_reference_info()
+        for tree_path, (branch_location, file_id) in viewitems(reference_dict):
+            branch_location = urlutils.rebase_url(branch_location,
+                                                  old_base, new_base)
+            target_reference_dict.setdefault(
+                tree_path, (branch_location, file_id))
+        self.target._set_all_reference_info(target_reference_dict)
+
 
 InterBranch.register_optimiser(GenericInterBranch)

=== modified file 'breezy/builtins.py'
--- breezy/builtins.py	2020-01-12 14:46:56 +0000
+++ breezy/builtins.py	2020-01-18 22:02:05 +0000
@@ -6723,36 +6723,34 @@
 
     takes_args = ['path?', 'location?']
     takes_options = [
+        'directory',
         Option('force-unversioned',
                help='Set reference even if path is not versioned.'),
         ]
 
-    def run(self, path=None, location=None, force_unversioned=False):
-        branchdir = '.'
-        if path is not None:
-            branchdir = path
+    def run(self, path=None, directory='.', location=None, force_unversioned=False):
         tree, branch, relpath = (
-            controldir.ControlDir.open_containing_tree_or_branch(branchdir))
-        if path is not None:
-            path = relpath
+            controldir.ControlDir.open_containing_tree_or_branch(directory))
         if tree is None:
             tree = branch.basis_tree()
         if path is None:
-            info = viewitems(branch._get_all_reference_info())
-            self._display_reference_info(tree, branch, info)
+            with tree.lock_read():
+                info = [
+                    (path, tree.get_reference_info(path, branch))
+                    for path in tree.iter_references()]
+                self._display_reference_info(tree, branch, info)
         else:
             if not tree.is_versioned(path) and not force_unversioned:
                 raise errors.NotVersionedError(path)
             if location is None:
-                info = [(path, branch.get_reference_info(path))]
+                info = [(path, tree.get_reference_info(path, branch))]
                 self._display_reference_info(tree, branch, info)
             else:
-                branch.set_reference_info(
-                    path, location, file_id=tree.path2id(path))
+                tree.set_reference_info(path, location)
 
     def _display_reference_info(self, tree, branch, info):
         ref_list = []
-        for path, (location, file_id) in info:
+        for path, location in info:
             ref_list.append((path, location))
         for path, location in sorted(ref_list):
             self.outf.write('%s %s\n' % (path, location))

=== modified file 'breezy/bzr/branch.py'
--- breezy/bzr/branch.py	2020-01-18 15:43:04 +0000
+++ breezy/bzr/branch.py	2020-01-19 03:22:04 +0000
@@ -450,16 +450,6 @@
             reconciler = BranchReconciler(self, thorough=thorough)
             return reconciler.reconcile()
 
-    def reference_parent(self, file_id, path, possible_transports=None):
-        """Return the parent branch for a tree-reference file_id
-
-        :param path: The path of the nested tree in the tree
-        :return: A branch associated with the nested tree
-        """
-        # FIXME should provide multiple branches, based on config
-        return Branch.open(self.controldir.root_transport.clone(path).base,
-                           possible_transports=possible_transports)
-
     def set_reference_info(self, file_id, branch_location, path=None):
         """Set the branch location to use for a tree reference."""
         raise errors.UnsupportedOperation(self.set_reference_info, self)
@@ -468,6 +458,19 @@
         """Get the tree_path and branch_location for a tree reference."""
         raise errors.UnsupportedOperation(self.get_reference_info, self)
 
+    def reference_parent(self, file_id, path, possible_transports=None):
+        """Return the parent branch for a tree-reference.
+
+        :param path: The path of the nested tree in the tree
+        :return: A branch associated with the nested tree
+        """
+        try:
+            return Branch.open_from_transport(
+                self.controldir.root_transport.clone(path),
+                possible_transports=possible_transports)
+        except errors.NotBranchError:
+            return None
+
 
 class BzrBranch8(BzrBranch):
     """A branch that stores tree-reference locations."""
@@ -540,7 +543,7 @@
         """
         s = BytesIO()
         writer = rio.RioWriter(s)
-        for tree_path, (branch_location, tree_path) in viewitems(info_dict):
+        for file_id, (branch_location, tree_path) in viewitems(info_dict):
             stanza = rio.Stanza(file_id=file_id,
                                 branch_location=branch_location)
             if tree_path is not None:
@@ -564,7 +567,7 @@
                     info_dict = {
                         s['file_id'].encode('ascii'): (
                             s['branch_location'],
-                            s['tree_path'])
+                            s['tree_path'] if 'tree_path' in s else None)
                         for s in stanzas}
             except errors.NoSuchFile:
                 info_dict = {}
@@ -600,10 +603,16 @@
         """
         branch_location = self.get_reference_info(file_id)[0]
         if branch_location is None:
-            return BzrBranch.reference_parent(self, path, possible_transports)
-        branch_location = urlutils.join(self.user_url, branch_location)
-        return Branch.open(branch_location,
-                           possible_transports=possible_transports)
+            try:
+                return Branch.open_from_transport(
+                    self.controldir.root_transport.clone(path),
+                    possible_transports=possible_transports)
+            except errors.NotBranchError:
+                return None
+        else:
+            branch_location = urlutils.join(self.user_url, branch_location)
+            return Branch.open(
+                branch_location, possible_transports=possible_transports)
 
     def set_push_location(self, location):
         """See Branch.set_push_location."""

=== modified file 'breezy/bzr/bzrdir.py'
--- breezy/bzr/bzrdir.py	2019-10-20 00:11:04 +0000
+++ breezy/bzr/bzrdir.py	2020-01-19 01:29:22 +0000
@@ -65,6 +65,7 @@
 from ..trace import (
     mutter,
     note,
+    warning,
     )
 
 from .. import (
@@ -453,22 +454,30 @@
             else:
                 wt = None
             if recurse == 'down':
-                basis = None
+                tree = None
                 if wt is not None:
-                    basis = wt.basis_tree()
+                    tree = wt
+                    basis = tree.basis_tree()
+                    stack.enter_context(basis.lock_read())
                 elif result_branch is not None:
-                    basis = result_branch.basis_tree()
+                    basis = tree = result_branch.basis_tree()
                 elif source_branch is not None:
-                    basis = source_branch.basis_tree()
-                if basis is not None:
-                    stack.enter_context(basis.lock_read())
-                    subtrees = basis.iter_references()
+                    basis = tree = source_branch.basis_tree()
+                if tree is not None:
+                    stack.enter_context(tree.lock_read())
+                    subtrees = tree.iter_references()
                 else:
                     subtrees = []
                 for path in subtrees:
                     target = urlutils.join(url, urlutils.escape(path))
-                    sublocation = source_branch.reference_parent(
-                        path, possible_transports=possible_transports)
+                    sublocation = tree.reference_parent(
+                        path, branch=result_branch,
+                        possible_transports=possible_transports)
+                    if sublocation is None:
+                        warning(
+                            'Ignoring nested tree %s, parent location unknown.',
+                            path)
+                        continue
                     sublocation.controldir.sprout(
                         target, basis.get_reference_revision(path),
                         force_new_repo=force_new_repo, recurse=recurse,

=== modified file 'breezy/bzr/inventorytree.py'
--- breezy/bzr/inventorytree.py	2020-01-11 02:39:17 +0000
+++ breezy/bzr/inventorytree.py	2020-01-19 03:22:04 +0000
@@ -23,6 +23,7 @@
 import re
 
 from .. import (
+    branch as _mod_branch,
     debug,
     errors,
     lazy_import,
@@ -199,7 +200,10 @@
                 trace.mutter_callsite(
                     2, "id2path with nested trees scales with tree size.")
             for path in self.iter_references():
-                subtree = self.get_nested_tree(path)
+                try:
+                    subtree = self.get_nested_tree(path)
+                except errors.NotBranchError:
+                    continue
                 try:
                     return osutils.pathjoin(path, subtree.id2path(file_id))
                 except errors.NoSuchId:
@@ -784,6 +788,23 @@
     def has_filename(self, filename):
         return bool(self.path2id(filename))
 
+    def reference_parent(self, path, branch=None, possible_transports=None):
+        if branch is not None:
+            file_id = self.path2id(path)
+            parent_url = branch.get_reference_info(file_id)[0]
+        else:
+            subdir = ControlDir.open_from_transport(
+                self._repository.user_transport.clone(path))
+            parent_url = subdir.open_branch().get_parent()
+        if parent_url is None:
+            return None
+        return _mod_branch.Branch.open(
+            parent_url,
+            possible_transports=possible_transports)
+
+    def get_reference_info(self, path, branch=None):
+        return branch.get_reference_info(self.path2id(path))[0]
+
     def list_files(self, include_root=False, from_dir=None, recursive=True,
                    recurse_nested=False):
         # The only files returned by this are those from the version
@@ -823,7 +844,8 @@
 
     def _get_nested_tree(self, path, file_id, reference_revision):
         # Just a guess..
-        subdir = ControlDir.open_from_transport(self._repository.user_transport.clone(path))
+        subdir = ControlDir.open_from_transport(
+            self._repository.user_transport.clone(path))
         subrepo = subdir.find_repository()
         try:
             revtree = subrepo.revision_tree(reference_revision)

=== modified file 'breezy/bzr/remote.py'
--- breezy/bzr/remote.py	2019-06-29 22:49:46 +0000
+++ breezy/bzr/remote.py	2020-01-18 22:02:05 +0000
@@ -3415,6 +3415,8 @@
                 return True
         return False
 
+    supports_reference_locations = False
+
 
 class RemoteBranchStore(_mod_config.IniFileStore):
     """Branch store which attempts to use HPSS calls to retrieve branch store.
@@ -4177,6 +4179,29 @@
             reconciler = BranchReconciler(self, thorough=thorough)
             return reconciler.reconcile()
 
+    def set_reference_info(self, tree_path, branch_location, file_id=None):
+        raise errors.UnsupportedOperation(self.set_reference_info, self)
+
+    def get_reference_info(self, tree_path):
+        raise errors.UnsupportedOperation(self.get_reference_info, self)
+
+    def _get_all_reference_info(self):
+        self._ensure_real()
+        return self._real_branch._get_all_reference_info()
+
+    def reference_parent(self, path, possible_transports=None):
+        """Return the parent branch for a tree-reference.
+
+        :param path: The path of the nested tree in the tree
+        :return: A branch associated with the nested tree
+        """
+        branch_location = self.get_reference_info(path)[0]
+        if branch_location is None:
+            return BzrBranch.reference_parent(self, path, possible_transports)
+        branch_location = urlutils.join(self.user_url, branch_location)
+        return Branch.open(branch_location,
+                           possible_transports=possible_transports)
+
 
 class RemoteConfig(object):
     """A Config that reads and writes from smart verbs.

=== modified file 'breezy/bzr/workingtree.py'
--- breezy/bzr/workingtree.py	2020-01-11 02:39:17 +0000
+++ breezy/bzr/workingtree.py	2020-01-19 03:22:04 +0000
@@ -1796,6 +1796,23 @@
                 else:
                     yield get_canonical_path(self, path, normalize)
 
+    def get_reference_info(self, path, branch=None):
+        file_id = self.path2id(path)
+        if file_id is None:
+            return None
+        return self.branch.get_reference_info(file_id)[0]
+
+    def set_reference_info(self, tree_path, branch_location):
+        file_id = self.path2id(tree_path)
+        if file_id is None:
+            raise errors.NoSuchFile(tree_path)
+        self.branch.set_reference_info(file_id, branch_location, tree_path)
+
+    def reference_parent(self, path, branch=None, possible_transports=None):
+        return self.branch.reference_parent(
+            self.path2id(path),
+            path, possible_transports=possible_transports)
+
 
 class WorkingTreeFormatMetaDir(bzrdir.BzrFormat, WorkingTreeFormat):
     """Base class for working trees that live in bzr meta directories."""

=== modified file 'breezy/bzr/workingtree_4.py'
--- breezy/bzr/workingtree_4.py	2020-01-11 22:33:18 +0000
+++ breezy/bzr/workingtree_4.py	2020-01-18 00:05:13 +0000
@@ -130,6 +130,9 @@
         self.views = self._make_views()
         # --- allow tests to select the dirstate iter_changes implementation
         self._iter_changes = dirstate._process_entry
+        self._repo_supports_tree_reference = getattr(
+            self._branch.repository._format, "supports_tree_reference",
+            False)
 
     def _add(self, files, ids, kinds):
         """See MutableTree._add."""
@@ -550,20 +553,21 @@
             # When the repo doesn't support references, we will have nothing to
             # return
             return
-        for key, tree_details in self.current_dirstate()._iter_entries():
-            if tree_details[0][0] in (b'a', b'r'):  # absent, relocated
-                # not relevant to the working tree
-                continue
-            if not key[1]:
-                # the root is not a reference.
-                continue
-            relpath = pathjoin(key[0].decode('utf8'), key[1].decode('utf8'))
-            try:
-                if self.kind(relpath) == 'tree-reference':
-                    yield relpath
-            except errors.NoSuchFile:
-                # path is missing on disk.
-                continue
+        with self.lock_read():
+            for key, tree_details in self.current_dirstate()._iter_entries():
+                if tree_details[0][0] in (b'a', b'r'):  # absent, relocated
+                    # not relevant to the working tree
+                    continue
+                if not key[1]:
+                    # the root is not a reference.
+                    continue
+                relpath = pathjoin(key[0].decode('utf8'), key[1].decode('utf8'))
+                try:
+                    if self.kind(relpath) == 'tree-reference':
+                        yield relpath
+                except errors.NoSuchFile:
+                    # path is missing on disk.
+                    continue
 
     def _observed_sha1(self, path, sha_and_stat):
         """See MutableTree._observed_sha1."""

=== modified file 'breezy/git/branch.py'
--- breezy/git/branch.py	2020-01-16 23:24:40 +0000
+++ breezy/git/branch.py	2020-01-18 16:15:37 +0000
@@ -22,6 +22,11 @@
 from io import BytesIO
 from collections import defaultdict
 
+from dulwich.config import (
+    ConfigFile as GitConfigFile,
+    parse_submodules,
+    )
+
 from dulwich.objects import (
     NotCommitError,
     ZERO_SHA,
@@ -809,17 +814,6 @@
         return GitMemoryTree(self, self.repository._git.object_store,
                              self.head)
 
-    def reference_parent(self, path, possible_transports=None):
-        """Return the parent branch for a tree-reference.
-
-        :param path: The path of the nested tree in the tree
-        :return: A branch associated with the nested tree
-        """
-        # FIXME should provide multiple branches, based on config
-        url = urlutils.strip_segment_parameters(self.user_url)
-        url = urlutils.join(url, path)
-        return branch.Branch.open(url, possible_transports=possible_transports)
-
 
 def _quick_lookup_revno(local_branch, remote_branch, revid):
     if not isinstance(revid, bytes):
@@ -980,6 +974,20 @@
             other_branch=self.source)
         return head, refs
 
+    def update_references(self, revid=None):
+        if revid is None:
+            revid = self.target.last_revision()
+        tree = self.target.repository.revision_tree(revid)
+        try:
+            with tree.get_file('.gitmodules') as f:
+                for path, url, section in parse_submodules(
+                        GitConfigFile.from_file(f)):
+                    self.target.set_reference_info(
+                        path.decode('utf-8'), url.decode('utf-8'),
+                        tree.path2id(path.decode('utf-8')))
+        except errors.NoSuchFile:
+            pass
+
     def _basic_pull(self, stop_revision, overwrite, run_hooks,
                     _override_hook_target, _hook_master):
         if overwrite is True:
@@ -1007,6 +1015,7 @@
                 result.tag_conflicts = tags_ret
             (result.new_revno, result.new_revid) = \
                 self.target.last_revision_info()
+            self.update_references(revid=result.new_revid)
             if _hook_master:
                 result.master_branch = _hook_master
                 result.local_branch = result.target_branch
@@ -1075,6 +1084,7 @@
             self.target.tags, "tags" in overwrite, ignore_master=True)
         (result.tag_updates, result.tag_conflicts) = tags_ret
         result.new_revno, result.new_revid = self.target.last_revision_info()
+        self.update_references(revid=result.new_revid)
         return result
 
 

=== modified file 'breezy/git/dir.py'
--- breezy/git/dir.py	2020-01-12 14:46:56 +0000
+++ breezy/git/dir.py	2020-01-18 16:14:28 +0000
@@ -207,8 +207,13 @@
                     subtrees = []
                 for path in subtrees:
                     target = urlutils.join(url, urlutils.escape(path))
-                    sublocation = source_branch.reference_parent(
+                    sublocation = wt.reference_parent(
                         path, possible_transports=possible_transports)
+                    if sublocation is None:
+                        trace.warning(
+                            'Ignoring nested tree %s, parent location unknown.',
+                            path)
+                        continue
                     sublocation.controldir.sprout(
                         target, basis.get_reference_revision(path),
                         force_new_repo=force_new_repo, recurse=recurse,

=== modified file 'breezy/git/fetch.py'
--- breezy/git/fetch.py	2019-10-28 01:38:39 +0000
+++ breezy/git/fetch.py	2020-01-18 00:08:16 +0000
@@ -55,6 +55,7 @@
 from ..bzr.testament import (
     StrictTestament3,
     )
+from ..tree import find_previous_path
 from ..tsort import (
     topo_sort,
     )
@@ -125,9 +126,8 @@
     # Check what revision we should store
     parent_keys = []
     for ptree in parent_bzr_trees:
-        try:
-            ppath = ptree.id2path(file_id)
-        except errors.NoSuchId:
+        ppath = find_previous_path(base_bzr_tree, ptree, decoded_path, file_id)
+        if ppath is None:
             continue
         pkind = ptree.kind(ppath)
         if (pkind == ie.kind and

=== modified file 'breezy/git/workingtree.py'
--- breezy/git/workingtree.py	2020-01-10 01:37:30 +0000
+++ breezy/git/workingtree.py	2020-01-19 01:29:22 +0000
@@ -25,6 +25,7 @@
 from dulwich.ignore import (
     IgnoreFilterManager,
     )
+from dulwich.config import ConfigFile as GitConfigFile
 from dulwich.file import GitFile, FileLocked
 from dulwich.index import (
     Index,
@@ -49,6 +50,7 @@
 import sys
 
 from .. import (
+    branch as _mod_branch,
     conflicts as _mod_conflicts,
     errors,
     controldir as _mod_controldir,
@@ -61,6 +63,7 @@
     trace,
     transport as _mod_transport,
     tree,
+    urlutils,
     workingtree,
     )
 from ..decorators import (
@@ -1240,7 +1243,7 @@
     def _read_submodule_head(self, path):
         return read_submodule_head(self.abspath(path))
 
-    def get_reference_revision(self, path):
+    def get_reference_revision(self, path, branch=None):
         hexsha = self._read_submodule_head(path)
         if hexsha is None:
             return _mod_revision.NULL_REVISION
@@ -1325,6 +1328,48 @@
                     new_parents = [revision_id]
                 tree.set_parent_ids(new_parents)
 
+    def reference_parent(self, path, possible_transports=None):
+        remote_url = self.get_reference_info(path)
+        if remote_url is None:
+            trace.warning("Unable to find submodule info for %s", path)
+            return None
+        return _mod_branch.Branch.open(remote_url, possible_transports=possible_transports)
+
+    def get_reference_info(self, path):
+        submodule_info = self._submodule_info()
+        info = submodule_info.get(path.encode('utf-8'))
+        if info is None:
+            return None
+        return info[0].decode('utf-8')
+
+    def set_reference_info(self, tree_path, branch_location):
+        path = self.abspath('.gitmodules')
+        try:
+            config = GitConfigFile.from_path(path)
+        except EnvironmentError as e:
+            if e.errno == errno.ENOENT:
+                config = GitConfigFile()
+            else:
+                raise
+        section = (b'submodule', tree_path.encode('utf-8'))
+        if branch_location is None:
+            try:
+                del config[section]
+            except KeyError:
+                pass
+        else:
+            branch_location = urlutils.join(
+                urlutils.strip_segment_parameters(self.branch.user_url),
+                branch_location)
+            config.set(
+                section,
+                b'path', tree_path.encode('utf-8'))
+            config.set(
+                section,
+                b'url', branch_location.encode('utf-8'))
+        config.write_to_path(path)
+        self.add('.gitmodules')
+
 
 class GitWorkingTreeFormat(workingtree.WorkingTreeFormat):
 

=== modified file 'breezy/merge.py'
--- breezy/merge.py	2019-10-13 17:31:55 +0000
+++ breezy/merge.py	2020-01-18 16:14:28 +0000
@@ -641,7 +641,7 @@
                     continue
                 sub_merge = Merger(sub_tree.branch, this_tree=sub_tree)
                 sub_merge.merge_type = self.merge_type
-                other_branch = self.other_branch.reference_parent(relpath)
+                other_branch = self.other_tree.reference_parent(relpath)
                 sub_merge.set_other_revision(other_revision, other_branch)
                 base_tree_path = _mod_tree.find_previous_path(
                     self.this_tree, self.base_tree, relpath)

=== modified file 'breezy/tests/blackbox/test_branch.py'
--- breezy/tests/blackbox/test_branch.py	2019-10-19 23:18:56 +0000
+++ breezy/tests/blackbox/test_branch.py	2020-01-19 03:22:04 +0000
@@ -367,6 +367,7 @@
         subtree.add(['a'])
         subtree.commit('add subtree contents')
         orig.add_reference(subtree)
+        orig.set_reference_info('subtree', subtree.branch.user_url)
         orig.commit('add subtree')
 
         self.run_bzr('branch source target')
@@ -376,6 +377,20 @@
         self.assertTreesEqual(orig, target)
         self.assertTreesEqual(subtree, target_subtree)
 
+    def test_branch_with_nested_trees_reference_unset(self):
+        orig = self.make_branch_and_tree('source', format='development-subtree')
+        subtree = self.make_branch_and_tree('source/subtree')
+        self.build_tree(['source/subtree/a'])
+        subtree.add(['a'])
+        subtree.commit('add subtree contents')
+        orig.add_reference(subtree)
+        orig.commit('add subtree')
+
+        self.run_bzr('branch source target')
+
+        target = WorkingTree.open('target')
+        self.assertRaises(errors.NotBranchError, WorkingTree.open, 'target/subtree')
+
     def test_branch_with_nested_trees_no_recurse(self):
         orig = self.make_branch_and_tree('source', format='development-subtree')
         subtree = self.make_branch_and_tree('source/subtree')

=== modified file 'breezy/tests/blackbox/test_reference.py'
--- breezy/tests/blackbox/test_reference.py	2018-11-16 11:37:47 +0000
+++ breezy/tests/blackbox/test_reference.py	2020-01-19 03:22:04 +0000
@@ -30,9 +30,12 @@
         return controldir.format_registry.make_controldir('development-subtree')
 
     def test_no_args_lists(self):
-        branch = self.make_branch('branch')
-        branch.set_reference_info('path', 'http://example.org', b'file-id')
-        branch.set_reference_info('lath', 'http://example.org/2', b'file-id2')
+        tree = self.make_branch_and_tree('branch')
+        branch = tree.branch
+        tree.add_reference(self.make_branch_and_tree('branch/path'))
+        tree.add_reference(self.make_branch_and_tree('branch/lath'))
+        tree.set_reference_info('path', 'http://example.org')
+        tree.set_reference_info('lath', 'http://example.org/2')
         out, err = self.run_bzr('reference', working_dir='branch')
         lines = out.splitlines()
         self.assertEqual('lath http://example.org/2', lines[0])
@@ -40,12 +43,10 @@
 
     def make_tree_with_reference(self):
         tree = self.make_branch_and_tree('tree')
-        self.build_tree(['tree/newpath'])
-        tree.add('newpath', b'file-id')
-        tree.branch.set_reference_info(
-            'newpath', 'http://example.org', b'file-id')
-        tree.branch.set_reference_info('lath', 'http://example.org/2',
-                                       b'file-id2')
+        subtree = self.make_branch_and_tree('tree/newpath')
+        tree.add_reference(subtree)
+        tree.set_reference_info('newpath', 'http://example.org')
+        tree.commit('add reference')
         return tree
 
     def test_uses_working_tree_location(self):
@@ -55,7 +56,6 @@
 
     def test_uses_basis_tree_location(self):
         tree = self.make_tree_with_reference()
-        tree.commit('add newpath')
         tree.controldir.destroy_workingtree()
         out, err = self.run_bzr('reference', working_dir='tree')
         self.assertContainsRe(out, 'newpath http://example.org\n')
@@ -67,17 +67,16 @@
 
     def test_one_arg_uses_containing_tree(self):
         tree = self.make_tree_with_reference()
-        out, err = self.run_bzr('reference tree/newpath')
+        out, err = self.run_bzr('reference -d tree newpath')
         self.assertEqual('newpath http://example.org\n', out)
 
     def test_two_args_sets(self):
         tree = self.make_branch_and_tree('tree')
         self.build_tree(['tree/file'])
-        tree.add('file', b'file-id')
-        out, err = self.run_bzr('reference tree/file http://example.org')
-        location, file_id = tree.branch.get_reference_info('file')
+        tree.add('file')
+        out, err = self.run_bzr('reference -d tree file http://example.org')
+        location = tree.get_reference_info('file')
         self.assertEqual('http://example.org', location)
-        self.assertEqual(b'file-id', file_id)
         self.assertEqual('', out)
         self.assertEqual('', err)
 
@@ -89,10 +88,11 @@
 
     def test_missing_file_forced(self):
         tree = self.make_branch_and_tree('tree')
+        tree.add_reference(self.make_branch_and_tree('tree/file'))
         out, err = self.run_bzr(
             'reference --force-unversioned file http://example.org',
             working_dir='tree')
-        location, file_id = tree.branch.get_reference_info('file')
+        location = tree.get_reference_info('file')
         self.assertEqual('http://example.org', location)
         self.assertEqual('', out)
         self.assertEqual('', err)

=== modified file 'breezy/tests/per_branch/test_branch.py'
--- breezy/tests/per_branch/test_branch.py	2020-01-12 15:16:27 +0000
+++ breezy/tests/per_branch/test_branch.py	2020-01-18 16:14:28 +0000
@@ -846,195 +846,6 @@
         self.assertLength(1, reopened.repository._fallback_repositories)
 
 
-class TestReferenceLocation(per_branch.TestCaseWithBranch):
-
-    def test_reference_parent(self):
-        tree = self.make_branch_and_tree('tree')
-        subtree = self.make_branch_and_tree('tree/subtree')
-        subtree.commit('a change')
-        try:
-            tree.add_reference(subtree)
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Tree cannot hold references.')
-        reference_parent = tree.branch.reference_parent(
-            urlutils.relative_url(
-                urlutils.split_segment_parameters(tree.branch.user_url)[0],
-                urlutils.split_segment_parameters(subtree.branch.user_url)[0]))
-        self.assertEqual(subtree.branch.user_url, reference_parent.user_url)
-
-    def test_reference_parent_accepts_possible_transports(self):
-        tree = self.make_branch_and_tree('tree')
-        subtree = self.make_branch_and_tree('tree/subtree')
-        subtree.commit('a change')
-        try:
-            tree.add_reference(subtree)
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Tree cannot hold references.')
-        reference_parent = tree.branch.reference_parent(
-            urlutils.relative_url(
-                urlutils.split_segment_parameters(tree.branch.user_url)[0],
-                urlutils.split_segment_parameters(subtree.branch.user_url)[0]),
-            possible_transports=[subtree.controldir.root_transport])
-
-    def test_get_reference_info(self):
-        branch = self.make_branch('branch')
-        try:
-            path, loc = branch.get_reference_info('file')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        self.assertIs(None, path)
-        self.assertIs(None, loc)
-
-    def test_set_reference_info(self):
-        branch = self.make_branch('branch')
-        try:
-            branch.set_reference_info('path/to/file', 'path/to/location',
-                                      b'file-id')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-
-    def test_set_get_reference_info(self):
-        branch = self.make_branch('branch')
-        try:
-            branch.set_reference_info('path/to/file',
-                                      'path/to/location', b'file-id')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        # Create a new instance to ensure storage is permanent
-        branch = _mod_branch.Branch.open('branch')
-        branch_location, file_id = branch.get_reference_info('path/to/file')
-        self.assertEqual('path/to/location', branch_location)
-
-    def test_set_null_reference_info(self):
-        branch = self.make_branch('branch')
-        try:
-            branch.set_reference_info('path/to/file',
-                                      'path/to/location', b'file-id')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        branch.set_reference_info('path/to/file', None, None)
-        branch_location, file_id = branch.get_reference_info('path/to/file')
-        self.assertIs(None, file_id)
-        self.assertIs(None, branch_location)
-
-    def test_set_null_reference_info_when_null(self):
-        branch = self.make_branch('branch')
-        try:
-            branch_location, file_id = branch.get_reference_info('file')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        self.assertIs(None, file_id)
-        self.assertIs(None, branch_location)
-        branch.set_reference_info('path/to/file', None, None)
-
-    def make_branch_with_reference(self, location, reference_location,
-                                   file_id='file-id'):
-        branch = self.make_branch(location)
-        try:
-            branch.set_reference_info('path/to/file',
-                                      reference_location, file_id)
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        return branch
-
-    def test_reference_parent_from_reference_info_(self):
-        referenced_branch = self.make_branch('reference_branch')
-        branch = self.make_branch_with_reference('branch',
-                                                 referenced_branch.base)
-        parent = branch.reference_parent('path/to/file')
-        self.assertEqual(parent.base, referenced_branch.base)
-
-    def test_branch_relative_reference_location(self):
-        branch = self.make_branch('branch')
-        try:
-            branch.set_reference_info('path/to/file',
-                                      '../reference_branch', b'file-id')
-        except errors.UnsupportedOperation:
-            raise tests.TestNotApplicable('Branch cannot hold references.')
-        referenced_branch = self.make_branch('reference_branch')
-        parent = branch.reference_parent('path/to/file')
-        self.assertEqual(parent.base, referenced_branch.base)
-
-    def test_sprout_copies_reference_location(self):
-        branch = self.make_branch_with_reference('branch', '../reference')
-        new_branch = branch.controldir.sprout('new-branch').open_branch()
-        self.assertEqual('../reference',
-                         new_branch.get_reference_info('path/to/file')[0])
-
-    def test_clone_copies_reference_location(self):
-        branch = self.make_branch_with_reference('branch', '../reference')
-        new_branch = branch.controldir.clone('new-branch').open_branch()
-        self.assertEqual('../reference',
-                         new_branch.get_reference_info('path/to/file')[0])
-
-    def test_copied_locations_are_rebased(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = branch.controldir.sprout(
-            'branch/new-branch').open_branch()
-        self.assertEqual('../reference',
-                         new_branch.get_reference_info('path/to/file')[0])
-
-    def test_update_references_retains_old_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = self.make_branch_with_reference(
-            'new_branch', 'reference', b'file-id2')
-        new_branch.update_references(branch)
-        self.assertEqual('reference',
-                         branch.get_reference_info('path/to/file')[0])
-
-    def test_update_references_retains_known_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = self.make_branch_with_reference(
-            'new_branch', 'reference2')
-        new_branch.update_references(branch)
-        self.assertEqual('reference',
-                         branch.get_reference_info('path/to/file')[0])
-
-    def test_update_references_skips_known_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = branch.controldir.sprout(
-            'branch/new-branch').open_branch()
-        new_branch.set_reference_info('../foo', '../foo', b'file-id')
-        new_branch.update_references(branch)
-        self.assertEqual('reference',
-                         branch.get_reference_info('path/to/file')[0])
-
-    def test_pull_updates_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = branch.controldir.sprout(
-            'branch/new-branch').open_branch()
-        new_branch.set_reference_info('../foo', '../foo', b'file-id2')
-        branch.pull(new_branch)
-        self.assertEqual('foo',
-                         branch.get_reference_info('../foo')[0])
-
-    def test_push_updates_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        new_branch = branch.controldir.sprout(
-            'branch/new-branch').open_branch()
-        new_branch.set_reference_info('../foo', '../foo', b'file-id2')
-        new_branch.push(branch)
-        self.assertEqual('foo',
-                         branch.get_reference_info('../foo')[0])
-
-    def test_merge_updates_references(self):
-        branch = self.make_branch_with_reference('branch', 'reference')
-        tree = self.make_branch_and_tree('tree')
-        tree.commit('foo')
-        branch.pull(tree.branch)
-        checkout = branch.create_checkout('checkout', lightweight=True)
-        checkout.commit('bar')
-        tree.lock_write()
-        self.addCleanup(tree.unlock)
-        merger = merge.Merger.from_revision_ids(tree,
-                                                branch.last_revision(),
-                                                other_branch=branch)
-        merger.merge_type = merge.Merge3Merger
-        merger.do_merge()
-        self.assertEqual('../branch/reference',
-                         tree.branch.get_reference_info('path/to/file')[0])
-
-
 class TestBranchControlComponent(per_branch.TestCaseWithBranch):
     """Branch implementations adequately implement ControlComponent."""
 

=== modified file 'breezy/tests/per_tree/test_tree.py'
--- breezy/tests/per_tree/test_tree.py	2019-09-27 02:48:17 +0000
+++ breezy/tests/per_tree/test_tree.py	2020-01-18 00:03:17 +0000
@@ -78,7 +78,7 @@
 class TestReference(TestCaseWithTree):
 
     def skip_if_no_reference(self, tree):
-        if not getattr(tree, 'supports_tree_reference', lambda: False)():
+        if not tree.supports_tree_reference():
             raise tests.TestNotApplicable('Tree references not supported')
 
     def create_nested(self):

=== modified file 'breezy/tests/per_workingtree/test_add_reference.py'
--- breezy/tests/per_workingtree/test_add_reference.py	2019-10-14 00:25:24 +0000
+++ breezy/tests/per_workingtree/test_add_reference.py	2020-01-18 00:05:13 +0000
@@ -61,6 +61,7 @@
                 self.assertEqual(
                     sub_tree.last_revision(),
                     tree.get_reference_revision('sub-tree'))
+        self.assertEqual(['sub-tree'], list(tree.iter_references()))
 
     def test_add_reference_same_root(self):
         tree = self.make_branch_and_tree('tree')

=== modified file 'breezy/tests/per_workingtree/test_workingtree.py'
--- breezy/tests/per_workingtree/test_workingtree.py	2019-10-15 19:05:47 +0000
+++ breezy/tests/per_workingtree/test_workingtree.py	2020-01-19 03:22:04 +0000
@@ -20,10 +20,11 @@
 import os
 
 from ... import (
-    branch,
+    branch as _mod_branch,
     config,
     controldir,
     errors,
+    merge,
     osutils,
     revision as _mod_revision,
     tests,
@@ -79,7 +80,7 @@
     def test_branch_builder(self):
         # Just a smoke test that we get a branch at the specified relpath
         builder = self.make_branch_builder('foobar')
-        br = branch.Branch.open(self.get_url('foobar'))
+        br = _mod_branch.Branch.open(self.get_url('foobar'))
 
     def test_list_files(self):
         tree = self.make_branch_and_tree('.')
@@ -299,7 +300,7 @@
     def test_initialize(self):
         # initialize should create a working tree and branch in an existing dir
         t = self.make_branch_and_tree('.')
-        b = branch.Branch.open('.')
+        b = _mod_branch.Branch.open('.')
         self.assertEqual(t.branch.base, b.base)
         t2 = WorkingTree.open('.')
         self.assertEqual(t.basedir, t2.basedir)
@@ -1329,3 +1330,217 @@
         self.assertSubset(
             [self.workingtree_format.supports_store_uncommitted],
             (True, False))
+
+
+class TestReferenceLocation(TestCaseWithWorkingTree):
+
+    def test_reference_parent(self):
+        tree = self.make_branch_and_tree('tree')
+        subtree = self.make_branch_and_tree('tree/subtree')
+        subtree.commit('a change')
+        try:
+            tree.add_reference(subtree)
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Tree cannot hold references.')
+        if not getattr(tree.branch._format, 'supports_reference_locations', False):
+            raise tests.TestNotApplicable('Branch cannot hold reference locations.')
+        tree.commit('Add reference.')
+        reference_parent = tree.reference_parent(
+            urlutils.relative_url(
+                urlutils.strip_segment_parameters(tree.branch.user_url),
+                urlutils.strip_segment_parameters(subtree.branch.user_url)))
+        self.assertEqual(subtree.branch.user_url, reference_parent.user_url)
+
+    def test_reference_parent_accepts_possible_transports(self):
+        tree = self.make_branch_and_tree('tree')
+        subtree = self.make_branch_and_tree('tree/subtree')
+        subtree.commit('a change')
+        try:
+            tree.add_reference(subtree)
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Tree cannot hold references.')
+        if not getattr(tree.branch._format, 'supports_reference_locations', False):
+            raise tests.TestNotApplicable('Branch cannot hold reference locations.')
+        tree.commit('Add reference')
+        reference_parent = tree.reference_parent(
+            urlutils.relative_url(
+                urlutils.strip_segment_parameters(tree.branch.user_url),
+                urlutils.strip_segment_parameters(subtree.branch.user_url)),
+            possible_transports=[subtree.controldir.root_transport])
+
+    def test_get_reference_info(self):
+        tree = self.make_branch_and_tree('branch')
+        try:
+            loc = tree.get_reference_info('file')
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Branch cannot hold references.')
+        self.assertIs(None, loc)
+
+    def test_set_reference_info(self):
+        self.make_tree_with_reference('branch', 'path/to/location')
+
+    def test_set_get_reference_info(self):
+        tree = self.make_tree_with_reference('branch', 'path/to/location')
+        # Create a new instance to ensure storage is permanent
+        tree = WorkingTree.open('branch')
+        branch_location = tree.get_reference_info('path/to/file')
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/path/to/location'),
+            urlutils.join(tree.branch.user_url, branch_location))
+
+    def test_set_null_reference_info(self):
+        tree = self.make_branch_and_tree('branch')
+        self.build_tree(['branch/file'])
+        tree.add(['file'])
+        try:
+            tree.set_reference_info('file', 'path/to/location')
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Branch cannot hold references.')
+        tree.set_reference_info('file', None)
+        branch_location = tree.get_reference_info('file')
+        self.assertIs(None, branch_location)
+
+    def test_set_null_reference_info_when_null(self):
+        tree = self.make_branch_and_tree('branch')
+        try:
+            branch_location = tree.get_reference_info('file')
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Branch cannot hold references.')
+        self.assertIs(None, branch_location)
+        self.build_tree(['branch/file'])
+        tree.add(['file'])
+        try:
+            tree.set_reference_info('file', None)
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Branch cannot hold references.')
+
+    def make_tree_with_reference(self, location, reference_location):
+        tree = self.make_branch_and_tree(location)
+        self.build_tree(
+            [os.path.join(location, name)
+             for name in ['path/', 'path/to/', 'path/to/file']])
+        tree.add(['path', 'path/to', 'path/to/file'])
+        try:
+            tree.set_reference_info('path/to/file', reference_location)
+        except errors.UnsupportedOperation:
+            raise tests.TestNotApplicable('Branch cannot hold references.')
+        tree.commit('commit reference')
+        return tree
+
+    def test_reference_parent_from_reference_info_(self):
+        referenced_branch = self.make_branch('reference_branch')
+        tree = self.make_tree_with_reference('branch', referenced_branch.base)
+        parent = tree.reference_parent('path/to/file')
+        self.assertEqual(parent.base, referenced_branch.base)
+
+    def test_branch_relative_reference_location(self):
+        tree = self.make_tree_with_reference('branch', '../reference_branch')
+        referenced_branch = self.make_branch('reference_branch')
+        parent = tree.reference_parent('path/to/file')
+        self.assertEqual(parent.base, referenced_branch.base)
+
+    def test_sprout_copies_reference_location(self):
+        tree = self.make_tree_with_reference('branch', '../reference')
+        new_tree = tree.branch.controldir.sprout('new-branch').open_workingtree()
+        self.assertEqual(
+            urlutils.local_path_to_url('reference'),
+            urlutils.join(urlutils.strip_segment_parameters(new_tree.branch.user_url),
+                          new_tree.get_reference_info('path/to/file')))
+
+    def test_clone_copies_reference_location(self):
+        tree = self.make_tree_with_reference('branch', '../reference')
+        new_tree = tree.controldir.clone('new-branch').open_workingtree()
+        self.assertEqual(
+            urlutils.local_path_to_url('reference'),
+            urlutils.join(urlutils.strip_segment_parameters(new_tree.branch.user_url),
+                          new_tree.get_reference_info('path/to/file')))
+
+    def test_copied_locations_are_rebased(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = tree.controldir.sprout(
+            'branch/new-branch').open_workingtree()
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/reference'),
+            urlutils.join(urlutils.strip_segment_parameters(new_tree.branch.user_url),
+                          new_tree.get_reference_info('path/to/file')))
+
+    def test_update_references_retains_old_references(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = self.make_tree_with_reference(
+            'new_branch', 'reference2')
+        new_tree.branch.update_references(tree.branch)
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/reference'),
+            urlutils.join(
+                urlutils.strip_segment_parameters(tree.branch.user_url),
+                tree.get_reference_info('path/to/file')))
+
+    def test_update_references_retains_known_references(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = self.make_tree_with_reference(
+            'new_branch', 'reference2')
+        new_tree.branch.update_references(tree.branch)
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/reference'),
+            urlutils.join(
+                urlutils.strip_segment_parameters(tree.branch.user_url),
+                tree.get_reference_info('path/to/file')))
+
+    def test_update_references_skips_known_references(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = tree.controldir.sprout(
+            'branch/new-branch').open_workingtree()
+        self.build_tree(['branch/new-branch/foo'])
+        new_tree.add('foo')
+        new_tree.set_reference_info('foo', '../foo')
+        new_tree.branch.update_references(tree.branch)
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/reference'),
+            urlutils.join(
+                urlutils.strip_segment_parameters(tree.branch.user_url),
+                tree.get_reference_info('path/to/file')))
+
+    def test_pull_updates_references(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = tree.controldir.sprout(
+            'branch/new-branch').open_workingtree()
+        self.build_tree(['branch/new-branch/foo'])
+        new_tree.add('foo')
+        new_tree.set_reference_info('foo', '../foo')
+        new_tree.commit('set reference')
+        tree.pull(new_tree.branch)
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/foo'),
+            urlutils.join(tree.branch.user_url, tree.get_reference_info('foo')))
+
+    def test_push_updates_references(self):
+        tree = self.make_tree_with_reference('branch', 'reference')
+        new_tree = tree.controldir.sprout(
+            'branch/new-branch').open_workingtree()
+        self.build_tree(['branch/new-branch/foo'])
+        new_tree.add(['foo'])
+        new_tree.set_reference_info('foo', '../foo')
+        new_tree.commit('add reference')
+        tree.pull(new_tree.branch)
+        tree.update()
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/foo'),
+            urlutils.join(tree.branch.user_url, tree.get_reference_info('foo')))
+
+    def test_merge_updates_references(self):
+        orig_tree = self.make_tree_with_reference('branch', 'reference')
+        tree = orig_tree.controldir.sprout('tree').open_workingtree()
+        tree.commit('foo')
+        orig_tree.pull(tree.branch)
+        checkout = orig_tree.branch.create_checkout('checkout', lightweight=True)
+        checkout.commit('bar')
+        tree.lock_write()
+        self.addCleanup(tree.unlock)
+        merger = merge.Merger.from_revision_ids(tree,
+                                                orig_tree.branch.last_revision(),
+                                                other_branch=orig_tree.branch)
+        merger.merge_type = merge.Merge3Merger
+        merger.do_merge()
+        self.assertEqual(
+            urlutils.local_path_to_url('branch/reference'),
+            urlutils.join(urlutils.strip_segment_parameters(tree.branch.user_url), tree.get_reference_info('path/to/file')))

=== modified file 'breezy/tests/test_branch.py'
--- breezy/tests/test_branch.py	2019-06-16 01:03:51 +0000
+++ breezy/tests/test_branch.py	2020-01-19 03:22:04 +0000
@@ -327,7 +327,7 @@
         self.assertPathDoesNotExist('a/.bzr/branch/bound')
         self.assertEqual('ftp://example.com', branch.get_bound_location())
 
-    def do_checkout_test(self, lightweight=False):
+    def do_checkout_test(self, lightweight):
         tree = self.make_branch_and_tree('source',
                                          format=self.get_format_name_subtree())
         subtree = self.make_branch_and_tree('source/subtree',
@@ -339,7 +339,9 @@
         subsubtree.add('file')
         subtree.add('file')
         subtree.add_reference(subsubtree)
+        subtree.set_reference_info('subsubtree', subsubtree.branch.user_url)
         tree.add_reference(subtree)
+        tree.set_reference_info('subtree', subtree.branch.user_url)
         tree.commit('a revision')
         subtree.commit('a subtree file')
         subsubtree.commit('a subsubtree file')
@@ -355,7 +357,7 @@
             self.assertEndsWith(subbranch.base, 'target/subtree/subsubtree/')
 
     def test_checkout_with_references(self):
-        self.do_checkout_test()
+        self.do_checkout_test(lightweight=False)
 
     def test_light_checkout_with_references(self):
         self.do_checkout_test(lightweight=True)
@@ -489,11 +491,11 @@
     def test_reference_info_caches_cleared(self):
         branch = self.make_branch('branch')
         with branch.lock_write():
-            branch.set_reference_info('path2', 'location2', b'file-id')
+            branch.set_reference_info(b'file-id', 'location2', 'path2')
         doppelganger = _mod_branch.Branch.open('branch')
-        doppelganger.set_reference_info('path3', 'location3', b'file-id')
-        self.assertEqual(('location3', b'file-id'),
-                         branch.get_reference_info('path3'))
+        doppelganger.set_reference_info(b'file-id', 'location3', 'path3')
+        self.assertEqual(('location3', 'path3'),
+                         branch.get_reference_info(b'file-id'))
 
     def _recordParentMapCalls(self, repo):
         self._parent_map_calls = []

=== modified file 'breezy/tests/test_bzrdir.py'
--- breezy/tests/test_bzrdir.py	2020-01-11 02:39:17 +0000
+++ breezy/tests/test_bzrdir.py	2020-01-19 03:22:04 +0000
@@ -842,12 +842,11 @@
         self.assertRaises(NotBranchError, bzrdir.BzrDir.open_from_transport, t)
 
     def test_sprout_recursive(self):
-        tree = self.make_branch_and_tree('tree1',
-                                         format='development-subtree')
-        sub_tree = self.make_branch_and_tree('tree1/subtree',
-                                             format='development-subtree')
+        tree = self.make_branch_and_tree('tree1')
+        sub_tree = self.make_branch_and_tree('tree1/subtree')
         sub_tree.set_root_id(b'subtree-root')
         tree.add_reference(sub_tree)
+        tree.set_reference_info('subtree', sub_tree.branch.user_url)
         self.build_tree(['tree1/subtree/file'])
         sub_tree.add('file')
         tree.commit('Initial commit')
@@ -872,12 +871,12 @@
         sub_tree = self.make_branch_and_tree('tree1/subtree',
                                              format='development-subtree')
         tree.add_reference(sub_tree)
+        tree.set_reference_info('subtree', sub_tree.branch.user_url)
         self.build_tree(['tree1/subtree/file'])
         sub_tree.add('file')
         tree.commit('Initial commit')
         # The following line force the orhaning to reveal bug #634470
-        tree.branch.get_config_stack().set(
-            'transform.orphan_policy', 'move')
+        tree.branch.get_config_stack().set('transform.orphan_policy', 'move')
         tree.controldir.destroy_workingtree()
         # FIXME: subtree/.bzr is left here which allows the test to pass (or
         # fail :-( ) -- vila 20100909

=== modified file 'breezy/tests/test_reconfigure.py'
--- breezy/tests/test_reconfigure.py	2018-11-12 01:41:38 +0000
+++ breezy/tests/test_reconfigure.py	2020-01-19 03:22:04 +0000
@@ -461,10 +461,10 @@
         format = controldir.format_registry.make_controldir('1.9')
         format.set_branch_format(_mod_bzrbranch.BzrBranchFormat8())
         tree = self.make_branch_and_tree('tree', format=format)
-        tree.branch.set_reference_info('path', '../location', 'file_id')
+        tree.branch.set_reference_info(b'file_id', '../location', 'path')
         checkout = tree.branch.create_checkout('checkout', lightweight=True)
         reconfiguration = reconfigure.Reconfigure.to_tree(checkout.controldir)
         reconfiguration.apply()
         checkout_branch = checkout.controldir.open_branch()
-        self.assertEqual(('../location', b'file_id'),
-                         checkout_branch.get_reference_info('path'))
+        self.assertEqual(('../location', 'path'),
+                         checkout_branch.get_reference_info(b'file_id'))

=== modified file 'breezy/transform.py'
--- breezy/transform.py	2020-01-10 01:40:18 +0000
+++ breezy/transform.py	2020-01-18 00:03:58 +0000
@@ -2036,6 +2036,11 @@
         self._iter_changes_cache = dict((c.file_id, c) for c in
                                         self._transform.iter_changes())
 
+    def supports_tree_reference(self):
+        # TODO(jelmer): Support tree references in _PreviewTree.
+        # return self._transform._tree.supports_tree_reference()
+        return False
+
     def _content_change(self, file_id):
         """Return True if the content of this file changed"""
         changes = self._iter_changes_cache.get(file_id)
@@ -2342,6 +2347,15 @@
             with self.get_file(path) as fileobj:
                 return sha_file(fileobj)
 
+    def get_reference_revision(self, path):
+        trans_id = self._path2trans_id(path)
+        if trans_id is None:
+            raise errors.NoSuchFile(path)
+        reference_revision = self._transform._new_reference_revision.get(trans_id)
+        if reference_revision is None:
+            return self._transform._tree.get_reference_revision(path)
+        return reference_revision
+
     def is_executable(self, path):
         trans_id = self._path2trans_id(path)
         if trans_id is None:

=== modified file 'breezy/tree.py'
--- breezy/tree.py	2020-01-10 01:40:18 +0000
+++ breezy/tree.py	2020-01-18 22:02:05 +0000
@@ -428,7 +428,7 @@
         """
         raise NotImplementedError(self.path_content_summary)
 
-    def get_reference_revision(self, path):
+    def get_reference_revision(self, path, branch=None):
         raise NotImplementedError("Tree subclass %s must implement "
                                   "get_reference_revision"
                                   % self.__class__.__name__)

=== modified file 'breezy/workingtree.py'
--- breezy/workingtree.py	2020-01-11 17:50:28 +0000
+++ breezy/workingtree.py	2020-01-19 01:29:22 +0000
@@ -1325,6 +1325,15 @@
         with self.lock_read():
             return next(self.get_canonical_paths([path]))
 
+    def reference_parent(self, path, branch=None, possible_transports=None):
+        raise errors.UnsupportedOperation(self.reference_parent, self)
+
+    def get_reference_info(self, path, branch=None):
+        raise errors.UnsupportedOperation(self.get_reference_info, self)
+
+    def set_reference_info(self, tree_path, branch_location):
+        raise errors.UnsupportedOperation(self.set_reference_info, self)
+
 
 class WorkingTreeFormatRegistry(ControlComponentFormatRegistry):
     """Registry for working tree formats."""

